# OptionSpec

OptionSpec describes the specification of a named Option. They have a name,
descr(iption), a short-name, a typ(e), and a default value when they are not
required.

Options can be placed anywhere before or after commands, and can be thought
of as named arguments.


```pony
class val OptionSpec
```

## Constructors

### bool

Creates an Option with a Bool typed value that can be used like
  `--opt` or `-O` or `--opt=true` or `-O=true`
to yield an option value like
  `cmd.option("opt").bool() == true`.


```pony
new val bool(
  name': String val,
  descr': String val = seq,
  short': (U8 val | None val) = seq,
  default': (Bool val | None val) = seq)
: OptionSpec val^
```
#### Parameters

*   name': [String](builtin-String) val
*   descr': [String](builtin-String) val = seq
*   short': ([U8](builtin-U8) val | [None](builtin-None) val) = seq
*   default': ([Bool](builtin-Bool) val | [None](builtin-None) val) = seq

#### Returns

* [OptionSpec](cli-OptionSpec) val^

---

### string

Creates an Option with a String typed value that can be used like
  `--file=dir/filename` or `-F=dir/filename` or `-Fdir/filename`
to yield an option value like
  `cmd.option("file").string() == "dir/filename"`.


```pony
new val string(
  name': String val,
  descr': String val = seq,
  short': (U8 val | None val) = seq,
  default': (String val | None val) = seq)
: OptionSpec val^
```
#### Parameters

*   name': [String](builtin-String) val
*   descr': [String](builtin-String) val = seq
*   short': ([U8](builtin-U8) val | [None](builtin-None) val) = seq
*   default': ([String](builtin-String) val | [None](builtin-None) val) = seq

#### Returns

* [OptionSpec](cli-OptionSpec) val^

---

### i64

Creates an Option with an I64 typed value that can be used like
  `--count=42 -C=42`
to yield an option value like
  `cmd.option("count").i64() == I64(64)`.


```pony
new val i64(
  name': String val,
  descr': String val = seq,
  short': (U8 val | None val) = seq,
  default': (I64 val | None val) = seq)
: OptionSpec val^
```
#### Parameters

*   name': [String](builtin-String) val
*   descr': [String](builtin-String) val = seq
*   short': ([U8](builtin-U8) val | [None](builtin-None) val) = seq
*   default': ([I64](builtin-I64) val | [None](builtin-None) val) = seq

#### Returns

* [OptionSpec](cli-OptionSpec) val^

---

### f64

Creates an Option with a F64 typed value that can be used like
  `--ratio=1.039` or `-R=1.039`
to yield an option value like
  `cmd.option("ratio").f64() == F64(1.039)`.


```pony
new val f64(
  name': String val,
  descr': String val = seq,
  short': (U8 val | None val) = seq,
  default': (F64 val | None val) = seq)
: OptionSpec val^
```
#### Parameters

*   name': [String](builtin-String) val
*   descr': [String](builtin-String) val = seq
*   short': ([U8](builtin-U8) val | [None](builtin-None) val) = seq
*   default': ([F64](builtin-F64) val | [None](builtin-None) val) = seq

#### Returns

* [OptionSpec](cli-OptionSpec) val^

---

### string_seq

Creates an Option with a ReadSeq[String] typed value that can be used like
  `--files=file1 --files=files2 --files=files2`
to yield a sequence of three strings equivalent to
  `cmd.option("ratio").string_seq() (equiv) ["file1"; "file2"; "file3"]`.


```pony
new val string_seq(
  name': String val,
  descr': String val = seq,
  short': (U8 val | None val) = seq)
: OptionSpec val^
```
#### Parameters

*   name': [String](builtin-String) val
*   descr': [String](builtin-String) val = seq
*   short': ([U8](builtin-U8) val | [None](builtin-None) val) = seq

#### Returns

* [OptionSpec](cli-OptionSpec) val^

---

## Public Functions

### name

Returns the name of this option.


```pony
fun box name()
: String val
```

#### Returns

* [String](builtin-String) val

---

### descr

Returns the description for this option.


```pony
fun box descr()
: String val
```

#### Returns

* [String](builtin-String) val

---

### required

Returns true iff this option is required to be present in the command line.


```pony
fun box required()
: Bool val
```

#### Returns

* [Bool](builtin-Bool) val

---

### help_string

Returns a formated help string for this option.


```pony
fun box help_string()
: String val
```

#### Returns

* [String](builtin-String) val

---

### deb_string

```pony
fun box deb_string()
: String val
```

#### Returns

* [String](builtin-String) val

---

## Private Functions

### _init

```pony
fun tag _init(
  typ': _ValueType val,
  default': (Bool val | String val | I64 val | 
    F64 val | _StringSeq val | None val))
: (_ValueType val , (Bool val | String val | I64 val | 
    F64 val | _StringSeq val) , Bool val)
```
#### Parameters

*   typ': [_ValueType](cli-_ValueType) val
*   default': ([Bool](builtin-Bool) val | [String](builtin-String) val | [I64](builtin-I64) val | 
    [F64](builtin-F64) val | [_StringSeq](cli-_StringSeq) val | [None](builtin-None) val)

#### Returns

* ([_ValueType](cli-_ValueType) val , ([Bool](builtin-Bool) val | [String](builtin-String) val | [I64](builtin-I64) val | 
    [F64](builtin-F64) val | [_StringSeq](cli-_StringSeq) val) , [Bool](builtin-Bool) val)

---

### _typ_p

```pony
fun box _typ_p()
: _ValueType val
```

#### Returns

* [_ValueType](cli-_ValueType) val

---

### _default_p

```pony
fun box _default_p()
: (Bool val | String val | I64 val | 
    F64 val | _StringSeq val)
```

#### Returns

* ([Bool](builtin-Bool) val | [String](builtin-String) val | [I64](builtin-I64) val | 
    [F64](builtin-F64) val | [_StringSeq](cli-_StringSeq) val)

---

### _requires_arg

```pony
fun box _requires_arg()
: Bool val
```

#### Returns

* [Bool](builtin-Bool) val

---

### _default_arg

```pony
fun box _default_arg()
: (Bool val | String val | I64 val | 
    F64 val | _StringSeq val)
```

#### Returns

* ([Bool](builtin-Bool) val | [String](builtin-String) val | [I64](builtin-I64) val | 
    [F64](builtin-F64) val | [_StringSeq](cli-_StringSeq) val)

---

### _has_short

```pony
fun box _has_short(
  sh: U8 val)
: Bool val
```
#### Parameters

*   sh: [U8](builtin-U8) val

#### Returns

* [Bool](builtin-Bool) val

---

